/**
 * 
 */
package com.wieseke.cptk.common.viewer;

import java.awt.Graphics2D;
import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.commands.operations.ObjectUndoContext;
import org.eclipse.jface.util.IPropertyChangeListener;
import org.eclipse.jface.util.PropertyChangeEvent;
import org.eclipse.jface.viewers.ISelection;
import org.eclipse.jface.viewers.ISelectionChangedListener;
import org.eclipse.jface.viewers.ISelectionProvider;
import org.eclipse.jface.viewers.SelectionChangedEvent;
import org.eclipse.jface.viewers.StructuredSelection;
import org.eclipse.swt.SWT;
import org.eclipse.swt.SWTException;
import org.eclipse.swt.custom.ScrolledComposite;
import org.eclipse.swt.events.KeyEvent;
import org.eclipse.swt.events.KeyListener;
import org.eclipse.swt.graphics.Color;
import org.eclipse.swt.graphics.Font;
import org.eclipse.swt.graphics.GC;
import org.eclipse.swt.graphics.Image;
import org.eclipse.swt.graphics.ImageData;
import org.eclipse.swt.graphics.Point;
import org.eclipse.swt.graphics.Rectangle;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.printing.PrintDialog;
import org.eclipse.swt.printing.Printer;
import org.eclipse.swt.printing.PrinterData;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Menu;
import org.eclipse.ui.PlatformUI;

import com.wieseke.cptk.common.api.ICophylogeny;
import com.wieseke.cptk.common.api.INode;
import com.wieseke.cptk.common.api.ITreeArrangement;
import com.wieseke.cptk.common.constants.ViewerConstants;
import com.wieseke.cptk.common.selection.CophylogenyViewerSelection;
import com.wieseke.cptk.common.util.Scaling;

/**
 * @author Nicolas Wieseke
 * 
 */
public abstract class CophylogenyViewer extends ScrolledComposite implements
		ISelectionProvider {

	/**
	 * the cophylogeny
	 */
	protected ICophylogeny cophylogeny;
	/**
	 * the container control for the node controls
	 */
	protected Composite container;

	/**
	 * the context menu for the node controls
	 */
	private Menu nodeContextMenu;

	/**
	 * the registered selection listeners
	 */
	private List<ISelectionChangedListener> listeners;
	/**
	 * the current selection
	 */
	private ISelection selection;

	/**
	 * list of host node controls
	 */
	protected List<NodeControl> hostNodesControls;
	/**
	 * list of parasite node controls
	 */
	protected List<NodeControl> parasiteNodesControls;

	/**
	 * the scale factor
	 */
	protected double scaleFactor;

	/**
	 * context for undo/redo operations in this viewer
	 */
	private IUndoContext undoContext;

	/**
	 * flag indicating if the content is unsaved
	 */
	private boolean dirty;

	/**
	 * flag indicating if the child nodes should retain there positions when the
	 * parent node is moved
	 */
	private boolean adjustChildNodes;
	/**
	 * flag indicating if the host node labels should be drawn
	 */
	protected boolean showHostLabels;
	/**
	 * flag indicating if the parasite node labels should be drawn
	 */
	protected boolean showParasiteLabels;
	/**
	 * flag indicating if the rank node labels should be drawn
	 */
	protected boolean showRankLabels;

	/**
	 * the width of the host lines
	 */
	protected int hostLineWidth;
	/**
	 * the width of the parasite lines
	 */
	protected int parasiteLineWidth;
	/**
	 * the color of the host lines
	 */
	protected Color hostLineColor;
	/**
	 * the color of the parasite lines
	 */
	protected Color parasiteLineColor;
	/**
	 * the font of the host names
	 */
	protected Font hostFont;
	/**
	 * the font of the parasite names
	 */
	protected Font parasiteFont;
	/**
	 * the color of the host font
	 */
	protected Color hostFontColor;
	/**
	 * the color of the parasite font
	 */
	protected Color parasiteFontColor;

	/**
	 * the name of the input used for printing
	 */
	private String name;

	private IPropertyChangeListener preferencesChangeListener;
	private List<IPropertyChangeListener> propertyChangeListeners;

	/**
	 * default constructor
	 * 
	 * @param parent
	 * @param style
	 * @param cophylogeny
	 */
	public CophylogenyViewer(Composite parent, ICophylogeny cophylogeny,
			String name) {
		super(parent, SWT.H_SCROLL | SWT.V_SCROLL);

		this.cophylogeny = cophylogeny;
		this.undoContext = new ObjectUndoContext(this);
		this.dirty = false;

		this.hostNodesControls = new ArrayList<NodeControl>();
		this.parasiteNodesControls = new ArrayList<NodeControl>();

		this.scaleFactor = 1d;

		this.adjustChildNodes = false;

		this.showHostLabels = true;
		this.showParasiteLabels = true;

		this.name = name;

		this.listeners = new ArrayList<ISelectionChangedListener>();

		this.propertyChangeListeners = new ArrayList<IPropertyChangeListener>();

		this.preferencesChangeListener = new IPropertyChangeListener() {
			
			public void propertyChange(PropertyChangeEvent event) {
				try {
					updatePreferences(event);
					CophylogenyViewer.this.redraw();
				} catch (SWTException e) {
					PlatformUI.getPreferenceStore()
							.removePropertyChangeListener(this);
				}
			}
		};
		PlatformUI.getPreferenceStore().addPropertyChangeListener(
				this.preferencesChangeListener);

		// add the key listener
		this.addKeyListener(new KeyListener() {
			
			public void keyPressed(KeyEvent e) {
				if (e.character == '+') {
					zoomIn();
				} else if (e.character == '-') {
					zoomOut();
				} else if (e.keyCode == SWT.ARROW_UP) {
					Point oldOrigin = CophylogenyViewer.this.getOrigin();
					CophylogenyViewer.this.setOrigin(oldOrigin.x, Math.max(0,
							oldOrigin.y - 10));
				} else if (e.keyCode == SWT.ARROW_DOWN) {
					Point oldOrigin = CophylogenyViewer.this.getOrigin();
					CophylogenyViewer.this.setOrigin(oldOrigin.x, Math.min(
							CophylogenyViewer.this.container.getSize().y,
							oldOrigin.y + 10));
				} else if (e.keyCode == SWT.ARROW_LEFT) {
					Point oldOrigin = CophylogenyViewer.this.getOrigin();
					CophylogenyViewer.this.setOrigin(Math.max(0,
							oldOrigin.x - 10), oldOrigin.y);
				} else if (e.keyCode == SWT.ARROW_RIGHT) {
					Point oldOrigin = CophylogenyViewer.this.getOrigin();
					CophylogenyViewer.this.setOrigin(Math.min(
							CophylogenyViewer.this.container.getSize().y,
							oldOrigin.x + 10), oldOrigin.y);
				}
			}

			
			public void keyReleased(KeyEvent e) {
			}
		});

		FillLayout layout = new FillLayout(SWT.HORIZONTAL);
		this.setLayout(layout);
		this.setBackground(new Color(Display.getDefault(), 255, 255, 255));

		// Create a container composite to hold the controls
		this.container = new Composite(this, SWT.TRANSPARENT);
		this.container.setLayout(null);

		this.setContent(container);

		this.setExpandHorizontal(true);
		this.setExpandVertical(true);

		this.getVerticalBar().setIncrement(10);

		init();
	}

	/**
	 * initialize the viewer
	 */
	protected abstract void init();

	/**
	 * 
	 * 
	 * @param event
	 */
	protected abstract void updatePreferences(PropertyChangeEvent event);

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.jface.viewers.ISelectionProvider#addSelectionChangedListener
	 * (org.eclipse.jface.viewers.ISelectionChangedListener)
	 */
	
	public void addSelectionChangedListener(ISelectionChangedListener listener) {
		if (this.listeners == null)
			this.listeners = new ArrayList<ISelectionChangedListener>();
		this.listeners.add(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.jface.viewers.ISelectionProvider#getSelection()
	 */
	
	public ISelection getSelection() {
		return (this.selection == null) ? StructuredSelection.EMPTY
				: this.selection;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.jface.viewers.ISelectionProvider#removeSelectionChangedListener
	 * (org.eclipse.jface.viewers.ISelectionChangedListener)
	 */
	
	public void removeSelectionChangedListener(
			ISelectionChangedListener listener) {
		if (this.listeners != null)
			this.listeners.remove(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.jface.viewers.ISelectionProvider#setSelection(org.eclipse
	 * .jface.viewers.ISelection)
	 */
	
	public void setSelection(ISelection selection) {
		SelectionChangedEvent event = new SelectionChangedEvent(this, selection);
		this.selection = selection;
		for (ISelectionChangedListener listener : this.listeners) {
			listener.selectionChanged(event);
		}
	}

	/**
	 * @param propertyName
	 * @param oldValue
	 * @param newValue
	 */
	protected void firePropertyChange(final String propertyName,
			final Object oldValue, final Object newValue) {
		PropertyChangeEvent event = new PropertyChangeEvent(this, propertyName,
				oldValue, newValue);
		for (IPropertyChangeListener listener : this.propertyChangeListeners) {
			listener.propertyChange(event);
		}
	}

	/**
	 * add a PropertyChangeListener
	 * 
	 * @param listener
	 */
	public void addPropertyChangeListener(final IPropertyChangeListener listener) {
		this.propertyChangeListeners.add(listener);
	}

	/**
	 * remove a PropertyChangeListener
	 * 
	 * @param listener
	 */
	public void removePropertyChangeListener(IPropertyChangeListener listener) {
		this.propertyChangeListeners.remove(listener);
	}

	/**
	 * @return the nodeContextMenu
	 */
	public Menu getNodeContextMenu() {
		return nodeContextMenu;
	}

	/**
	 * @param nodeContextMenu
	 *            the nodeContextMenu to set
	 */
	public void setNodeContextMenu(Menu nodeContextMenu) {
		this.nodeContextMenu = nodeContextMenu;
		for (NodeControl node : this.hostNodesControls) {
			node.setMenu(this.nodeContextMenu);
		}
		for (NodeControl node : this.parasiteNodesControls) {
			node.setMenu(this.nodeContextMenu);
		}
	}

	/**
	 * @param menu
	 *            the menu to set
	 */
	public void setViewerMenu(Menu menu) {
		this.container.setMenu(menu);
	}

	/**
	 * @return the cophylogeny
	 */
	public ICophylogeny getCophylogeny() {
		return cophylogeny;
	}

	/**
	 * @param cophylogeny
	 *            the cophylogeny to set
	 */
	public void setCophylogeny(ICophylogeny cophylogeny) {
		this.cophylogeny = cophylogeny;
	}

	/**
	 * @return the undoContext
	 */
	public IUndoContext getUndoContext() {
		return undoContext;
	}

	/**
	 * @return the dirty
	 */
	public boolean isDirty() {
		return dirty;
	}

	/**
	 * @param dirty
	 *            the dirty to set
	 */
	public void setDirty(boolean dirty) {
		if (this.dirty != dirty) {
			this.dirty = dirty;
			firePropertyChange(ViewerConstants.PROPERTY_DIRTY, !dirty, dirty);
		}
	}

	/**
	 * calculates the scaled size used by the specified tree
	 * 
	 * @return the rectangle with the size
	 */
	protected Rectangle getScaledTreeSize(int tree) {
		int minSizeX = Integer.MAX_VALUE;
		int maxSizeX = Integer.MIN_VALUE;
		int minSizeY = Integer.MAX_VALUE;
		int maxSizeY = Integer.MIN_VALUE;
		List<NodeControl> nodes = new ArrayList<NodeControl>();
		switch (tree) {
		case ViewerConstants.TREE_SELECTION_HOST:
			nodes = this.hostNodesControls;
			break;
		case ViewerConstants.TREE_SELECTION_PARASITE:
			nodes = this.parasiteNodesControls;
			break;
		case ViewerConstants.TREE_SELECTION_BOTH:
			Rectangle rectHost = getScaledTreeSize(ViewerConstants.TREE_SELECTION_HOST);
			Rectangle rectParasite = getScaledTreeSize(ViewerConstants.TREE_SELECTION_PARASITE);
			return new Rectangle(Math.min(rectHost.x, rectParasite.x), Math
					.min(rectHost.y, rectParasite.y), Math.max(rectHost.width,
					rectParasite.width), Math.max(rectHost.height,
					rectParasite.height));
		default:
			return null;
		}
		for (NodeControl node : nodes) {
			minSizeX = Math.min(minSizeX, scale(node.getPosX()));
			maxSizeX = Math.max(maxSizeX, scale(node.getPosX())
					+ scale(node.getNormalizedSize().x));
			minSizeY = Math.min(minSizeY, scale(node.getPosY()));
			maxSizeY = Math.max(maxSizeY, scale(node.getPosY())
					+ scale(node.getNormalizedSize().y));
		}
		return new Rectangle(minSizeX, minSizeY, maxSizeX, maxSizeY);
	}

	/**
	 * calculates the normalized size used by the specified tree
	 * 
	 * @return the rectangle with the normalized size
	 */
	public Rectangle getNormalizedTreeSize(int tree) {
		int minSizeX = Integer.MAX_VALUE;
		int maxSizeX = Integer.MIN_VALUE;
		int minSizeY = Integer.MAX_VALUE;
		int maxSizeY = Integer.MIN_VALUE;
		List<NodeControl> nodes = new ArrayList<NodeControl>();
		switch (tree) {
		case ViewerConstants.TREE_SELECTION_HOST:
			nodes = this.hostNodesControls;
			break;
		case ViewerConstants.TREE_SELECTION_PARASITE:
			nodes = this.parasiteNodesControls;
			break;
		case ViewerConstants.TREE_SELECTION_BOTH:
			Rectangle rectHost = getNormalizedTreeSize(ViewerConstants.TREE_SELECTION_HOST);
			Rectangle rectParasite = getNormalizedTreeSize(ViewerConstants.TREE_SELECTION_PARASITE);
			return new Rectangle(Math.min(rectHost.x, rectParasite.x), Math
					.min(rectHost.y, rectParasite.y), Math.max(rectHost.width,
					rectParasite.width), Math.max(rectHost.height,
					rectParasite.height));
		default:
			return null;
		}
		for (NodeControl node : nodes) {
			minSizeX = Math.min(minSizeX, node.getNode().getPosX());
			maxSizeX = Math.max(maxSizeX, node.getNode().getPosX() + 20);
			minSizeY = Math.min(minSizeY, node.getNode().getPosY());
			maxSizeY = Math.max(maxSizeY, node.getNode().getPosY() + 20);
		}
		return new Rectangle(minSizeX, minSizeY, maxSizeX, maxSizeY);
	}

	/**
	 * @param node
	 * @return the nodes Control
	 */
	protected NodeControl getNodeControlByINode(INode node) {
		for (NodeControl nodeControl : this.hostNodesControls) {
			if (nodeControl.getNode().equals(node))
				return nodeControl;
		}
		return null;
	}

	/**
	 * sets the min size of the scrolled composite
	 */
	public void resetMinSize() {
		Rectangle rect = getScaledTreeSize(ViewerConstants.TREE_SELECTION_BOTH);
		setMinSize(rect.width + 200, rect.height + 200);
	}

	/**
	 * initialize the visual representation objects of the cophylogeny
	 * 
	 * @param initNodes
	 * @param arrangement
	 */
	public abstract void initCophylogeny(boolean initNodes,
			ITreeArrangement arrangement);

	/**
	 * @return the tree arrangement
	 */
	public abstract ITreeArrangement getTreeArrangement();

	/**
	 * scale an integer to the current scale factor
	 * 
	 * @param value
	 * @return the scaled integer
	 */
	public int scale(int value) {
		return Scaling.convertToScaled(value, this.scaleFactor);
	}

	/**
	 * scale an integer to the current scale factor
	 * 
	 * @param value
	 * @return the scaled integer
	 */
	public int normalize(int value) {
		return Scaling.convertFromScaled(value, this.scaleFactor);
	}

	/**
	 * zoom the view to the specified scale factor
	 * 
	 * @param scaleFactor
	 */
	public void zoom(double scaleFactor) {
		this.scaleFactor = scaleFactor;// += 0.5d;
		container.setRedraw(false);
		for (NodeControl node : hostNodesControls) {
			node.setScaleFactor(scaleFactor);
			node.applyPosition();
		}
		for (NodeControl node : parasiteNodesControls) {
			node.setScaleFactor(scaleFactor);
			node.applyPosition();
		}
		this.resetMinSize();
		container.setRedraw(true);
	}

	/**
	 * zoom in the view
	 * 
	 * @param scaleFactor
	 */
	public void zoomIn() {
		this.container.setRedraw(false);
		zoom(this.scaleFactor * 1.1d);
		this.container.setRedraw(true);
	}

	/**
	 * zoom in the view
	 * 
	 * @param scaleFactor
	 */
	public void zoomOut() {
		this.container.setRedraw(false);
		zoom(this.scaleFactor / 1.1d);
		this.container.setRedraw(true);
	}

	/**
	 * Flips the selected tree horizontally.
	 */
	public abstract void flipHorizontal();

	/**
	 * Flips the selected tree vertically.
	 */
	public abstract void flipVertical();

	/**
	 * Rotate the selected tree to the left.
	 */
	public abstract void rotateLeft();

	/**
	 * Rotate the selected tree to the right.
	 */
	public abstract void rotateRight();

	/**
	 * @return the showHostLabels
	 */
	public boolean isShowHostLabels() {
		return showHostLabels;
	}

	/**
	 * @param showHostLabels
	 *            the showHostLabels to set
	 */
	public void setShowHostLabels(boolean showHostLabels) {
		boolean oldValue = this.showHostLabels;
		this.showHostLabels = showHostLabels;
		firePropertyChange(ViewerConstants.PROPERTY_SHOW_HOST_LABELS, oldValue,
				showHostLabels);
		redraw();
	}

	/**
	 * @return the showParasiteLabels
	 */
	public boolean isShowParasiteLabels() {
		return showParasiteLabels;
	}

	/**
	 * @param showParasiteLabels
	 *            the showParasiteLabels to set
	 */
	public void setShowParasiteLabels(boolean showParasiteLabels) {
		boolean oldValue = this.showParasiteLabels;
		this.showParasiteLabels = showParasiteLabels;
		firePropertyChange(ViewerConstants.PROPERTY_SHOW_PARASITE_LABELS,
				oldValue, showParasiteLabels);
		redraw();
	}

	/**
	 * @return the showRankLabels
	 */
	public boolean isShowRankLabels() {
		return showRankLabels;
	}

	/**
	 * @param showRankLabels
	 *            the showRankLabels to set
	 */
	public void setShowRankLabels(boolean showRankLabels) {
		boolean oldValue = this.showRankLabels;
		this.showRankLabels = showRankLabels;
		firePropertyChange(ViewerConstants.PROPERTY_SHOW_RANK_LABELS, oldValue,
				showRankLabels);
		redraw();
	}

	/**
	 * @return the adjustChildNodes
	 */
	public boolean isAdjustChildNode() {
		return adjustChildNodes;
	}

	/**
	 * @param adjustChildNodes
	 *            the adjustChildNodes to set
	 */
	public void setAdjustChildNodes(boolean adjustChildNodes) {
		boolean oldValue = this.adjustChildNodes;
		this.adjustChildNodes = adjustChildNodes;
		firePropertyChange(ViewerConstants.PROPERTY_ADJUST_CHILD_NODES,
				oldValue, adjustChildNodes);
	}

	/**
	 * print the content of the ScrolledComposite
	 */
	public void printContent() {
		Point size = this.getSize();
		final Image image = new Image(getDisplay(), size.x, size.y);
		GC gc = new GC(this);
		this.redraw();
		this.update();
		gc.copyArea(image, 0, 0);
		gc.dispose();

		// Get the ImageData and create a new printer Image from it
		ImageData imageData = image.getImageData();
		PrintDialog printDialog = new PrintDialog(getShell());
		// and open it
		PrinterData printerData = printDialog.open();
		// Check if OK was pressed
		if (printerData != null) {
			// Create new Printer instance
			Printer printer = new Printer(printerData);

			// Determine the bounds of the entire area of the printer
			Rectangle trim = printer.computeTrim(0, 0, 0, 0);
			Rectangle clientArea = printer.getClientArea();
			Rectangle area = new Rectangle(clientArea.x - trim.x, clientArea.y
					- trim.y, clientArea.width - trim.width, clientArea.height
					- trim.height);
			// Calculate the scale factor
			double scaleFactor = Math.min((double) area.width
					/ (double) imageData.width, (double) area.height
					/ (double) imageData.height);

			// Start the print job
			String filename = this.name
					.substring(this.name.lastIndexOf("/") + 1);
			if (printer.startJob(filename)) {
				if (printer.startPage()) {
					gc = new GC(printer);
					Image printerImage = new Image(printer, imageData);

					try {
						gc.drawImage(printerImage, 0, 0, imageData.width,
								imageData.height, -trim.x, -trim.y, (int) Math
										.round(scaleFactor * imageData.width),
								(int) Math
										.round(scaleFactor * imageData.height));
					} finally {
						// Clean up
						printerImage.dispose();
						gc.dispose();
						printer.endPage();
						printerImage.dispose();
					}

				}
			}
			// End the job and dispose the printer
			printer.endJob();
			printer.dispose();
		}
		image.dispose();
	}

	/**
	 * paints the content on the given g2d object
	 * 
	 * @param g2d
	 */
	public void paintContent(Graphics2D g2d) {
		for (NodeControl nodeControl : this.hostNodesControls) {
			nodeControl.paintControl(g2d);
		}
		for (NodeControl nodeControl : this.parasiteNodesControls) {
			nodeControl.paintControl(g2d);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.eclipse.swt.widgets.Composite#setFocus()
	 */
	
	public boolean setFocus() {
		setSelection(new CophylogenyViewerSelection(this));
		return super.setFocus();
	}

	
	public void dispose() {
		PlatformUI.getPreferenceStore().removePropertyChangeListener(
				this.preferencesChangeListener);
		super.dispose();
	}

	
	public void redraw() {
		this.container.changed(this.container.getChildren());
		this.container.redraw();
		super.redraw();
	}

}
