/**
 * 
 */
package com.wieseke.cptk.input.dao;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.TreeMap;

import com.wieseke.cptk.common.api.INode;
import com.wieseke.cptk.common.api.IPositionHostNode;
import com.wieseke.cptk.common.api.IPositionNode;
import com.wieseke.cptk.common.api.IPositionParasiteNode;
import com.wieseke.cptk.common.constants.CostsConstants;
import com.wieseke.cptk.common.constants.OptionsConstants;
import com.wieseke.cptk.common.dao.Cophylogeny;
import com.wieseke.cptk.common.dao.ParasiteRank;
import com.wieseke.cptk.common.dao.ReconstructionCosts;
import com.wieseke.cptk.common.util.Tupel;
import com.wieseke.cptk.reconstruction.dao.EventType;

/**
 * @author Nicolas Wieseke
 * 
 */
public class InputCophylogeny extends Cophylogeny {

	public static InputCophylogeny createNewInputCophylogeny(String filename) {
		InputCophylogeny cophylogeny = new InputCophylogeny(filename);
		InputHostNode hostRoot = new InputHostNode();
		hostRoot.setNumber(0);
		hostRoot.setLabel("HOST_ROOT");
		hostRoot.setLevel(0);
		hostRoot.setRank(0);
		hostRoot.setPosX(10);
		hostRoot.setPosY(10);
		cophylogeny.setHostRoot(hostRoot);
		InputParasiteNode parasiteRoot = new InputParasiteNode();
		parasiteRoot.setNumber(0);
		parasiteRoot.setLabel("PARASITE_ROOT");
		parasiteRoot.setLevel(0);
		parasiteRoot.setRank(new ParasiteRank(0, 0));
		parasiteRoot.setPosX(110);
		parasiteRoot.setPosY(10);
		cophylogeny.setParasiteRoot(parasiteRoot);
		cophylogeny.setCosts(createStandardCosts());
		cophylogeny.setOptions(createStandardOptions());
		return cophylogeny;
	}

	public static ReconstructionCosts createStandardCosts() {
		ReconstructionCosts costs = new ReconstructionCosts(
				CostsConstants.COSPECIATION_STDCOST,
				CostsConstants.SORTING_STDCOST,
				CostsConstants.DUPLICATION_STDCOST,
				CostsConstants.HOSTSWITCH_STDCOST);
		return costs;
	}

	public static Map<String, String> createStandardOptions() {
		Map<String, String> options = new HashMap<String, String>();
		options.put(OptionsConstants.HOSTSWITCH_ID,
				OptionsConstants.HOSTSWITCH_1_ID);
		options.put(OptionsConstants.SORTING_ID, OptionsConstants.SORTING_1_ID);
		options.put(OptionsConstants.DUPLICATION_ID,
				OptionsConstants.DUPLICATION_1_ID);
		options.put(OptionsConstants.TAKEOFF_ID, OptionsConstants.TAKEOFF_0_ID);
		options.put(OptionsConstants.RANK_ID, OptionsConstants.RANK_0_ID);
		options.put(OptionsConstants.ROOT_MAPPING_ID,
				OptionsConstants.ROOT_MAPPING_0_ID);
		options.put(OptionsConstants.PROBABILITY_COSTS_ID,
				OptionsConstants.PROBABILITY_COSTS_0_ID);
		options.put(OptionsConstants.FULL_HOSTSWITCH_ID,
				OptionsConstants.FULL_HOSTSWITCH_0_ID);
		options.put(OptionsConstants.LEAF_SPECIACION_COST_ID,
				OptionsConstants.LEAF_SPECIACION_COST_0_ID);
		options.put(OptionsConstants.CHECK_CHRONOLOGY_ID,
				OptionsConstants.CHECK_CHRONOLOGY_0_ID);
		options.put(OptionsConstants.AUTOMATIC_COSTS_ID,
				OptionsConstants.AUTOMATIC_COSTS_0_ID);
		options.put(OptionsConstants.AUTOMATIC_METHOD_ID,
				OptionsConstants.AUTOMATIC_METHOD_2_ID);
		options.put(OptionsConstants.RANDOM_CYCLES_ID, "5000");
		options.put(OptionsConstants.RANDOM_SEED_ID, "auto");
		// options.put(OptionsConstants.INTERPOLATION_THRESHOLD_ID, "0.01");
		// options.put(OptionsConstants.QUALITY_THRESHOLD_ID, "0.25");
		// options.put(OptionsConstants.RECURSION_DEPTH_ID, "3");
		return options;
	}

	/**
	 * the translation between the node numbers and their labels
	 */
	private Map<String, String> translations;
	/**
	 * the timezone information
	 */
	private Map<String, Tupel<Integer, Integer>> ranks;
	/**
	 * the association of parasite leafs to host leafs
	 */
	private Map<String, ArrayList<String>> phi;
	/**
	 * the reconstruction association of parasites to hosts
	 */
	private Map<String, ArrayList<String>> reconstruction;
	/**
	 * the reconstruction association of parasites to hosts
	 */
	private EventType reconstructionEvents;
	/**
	 * the positions of the nodes
	 */
	private Map<String, Tupel<Integer, Integer>> positions;

	/**
	 * default constructor
	 */
	public InputCophylogeny(String filename) {
		this.translations = new HashMap<String, String>();
		this.ranks = new HashMap<String, Tupel<Integer, Integer>>();
		this.phi = new HashMap<String, ArrayList<String>>();
		this.reconstruction = new HashMap<String, ArrayList<String>>();
		this.reconstructionEvents = new EventType(-1, -1, -1, -1);
		this.positions = new HashMap<String, Tupel<Integer, Integer>>();
		this.costs = createStandardCosts(); // new ReconstructionCosts();
		this.options = createStandardOptions(); // new HashMap<String,
												// String>();
		this.positionsApplied = false;
		setFilename(filename);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.wieseke.cptk.common.dao.Cophylogeny#getHostRoot()
	 */

	public InputHostNode getHostRoot() {
		return (InputHostNode) super.getHostRoot();
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see com.wieseke.cptk.common.dao.Cophylogeny#getParasiteRoot()
	 */

	public InputParasiteNode getParasiteRoot() {
		return (InputParasiteNode) super.getParasiteRoot();
	}

	/**
	 * @return the translation
	 */
	public Map<String, String> getTranslations() {
		return translations;
	}

	/**
	 * @param translation
	 *            the translation to set
	 */
	public void setTranslations(Map<String, String> translations) {
		this.translations = translations;
	}

	/**
	 * create a host tree from the tree data
	 * 
	 * @param tree
	 */
	public void createHostTree(IPositionHostNode hostRoot) {
		this.setHostRoot(hostRoot);
	}

	/**
	 * create a parasite tree from the tree data
	 * 
	 * @param tree
	 */
	public void createParasiteTree(IPositionParasiteNode parasiteRoot) {
		this.setParasiteRoot(parasiteRoot);
	}

	/**
	 * @return the ranks
	 */
	public Map<String, Tupel<Integer, Integer>> getRanks() {
		return ranks;
	}

	/**
	 * @param ranks
	 *            the ranks to set
	 */
	public void setRanks(Map<String, Tupel<Integer, Integer>> ranks) {
		this.ranks = ranks;
	}

	/**
	 * @return the phi
	 */
	public Map<String, ArrayList<String>> getPhi() {
		return phi;
	}

	/**
	 * @param phi
	 *            the phi to set
	 */
	public void setPhi(Map<String, ArrayList<String>> phi) {
		this.phi = phi;
	}

	/**
	 * @return the reconstruction
	 */
	public Map<String, ArrayList<String>> getReconstruction() {
		return reconstruction;
	}

	/**
	 * @param reconstruction
	 *            the reconstruction to set
	 */
	public void setReconstruction(Map<String, ArrayList<String>> reconstruction) {
		this.reconstruction = reconstruction;
	}

	public EventType getReconstructionEvents() {
		return reconstructionEvents;
	}

	public void setReconstructionEvents(EventType reconstructionEvents) {
		this.reconstructionEvents = reconstructionEvents;
	}

	/**
	 * @return the positions
	 */
	public Map<String, Tupel<Integer, Integer>> getPositions() {
		return positions;
	}

	/**
	 * @param positions
	 *            the positions to set
	 */
	public void setPositions(Map<String, Tupel<Integer, Integer>> positions) {
		this.positions = positions;
	}

	/**
	 * apply the translations from the nexus data to the host and parasite
	 * nodes.
	 */
	public void applyTranslations(IPositionNode node) {
		for (int index = 0; index < node.getChildren().size(); index++) {
			applyTranslations(node.getChildren().get(index));
		}
		String translation = this.translations.get(node.getLabel());
		if (translation != null)
			node.setLabel(translation);
	}

	/**
	 * apply the ranks from the nexus data to the host and parasite nodes.
	 */
	public void applyRanks(IPositionNode node) {
		for (int index = 0; index < node.getChildren().size(); index++) {
			applyRanks(node.getChildren().get(index));
		}
		Tupel<Integer, Integer> ranks = this.ranks.get(node.getLabel());
		if (ranks != null) {
			if (node instanceof IPositionHostNode) {
				((IPositionHostNode) node).setRank(ranks.getValue1());
			} else if (node instanceof IPositionParasiteNode) {
				((IPositionParasiteNode) node).setRank(new ParasiteRank(ranks
						.getValue1(), ranks.getValue2()));
			}
		}
	}

	/**
	 * apply the association from the nexus data to the parasite nodes.
	 */
	public void applyPhi(InputParasiteNode node) {
		for (int index = 0; index < node.getChildren().size(); index++) {
			applyPhi(node.getChildren().get(index));
		}
		ArrayList<String> phi = this.phi.get(node.getLabel());
		if (phi != null) {
			for (int index = 0; index < phi.size(); index++) {
				InputHostNode hostNode = (InputHostNode) getNodeByLabel(
						this.getHostRoot(), phi.get(index));
				if (hostNode != null)
					node.addAssociations(hostNode);
			}
		}
	}

	/**
	 * apply the reconstruction association from the nexus data to the parasite
	 * nodes.
	 */
	public void applyReconstruction(InputParasiteNode node) {
		for (int index = 0; index < node.getChildren().size(); index++) {
			applyReconstruction(node.getChildren().get(index));
		}
		ArrayList<String> reconstruction = this.reconstruction.get(node
				.getLabel());
		if (reconstruction != null) {
			for (int index = 0; index < reconstruction.size(); index++) {
				InputHostNode hostNode = (InputHostNode) getNodeByLabel(
						this.getHostRoot(), reconstruction.get(index));
				if (hostNode != null)
					node.addReconstructionAssociations(hostNode);
			}
		}
	}

	/**
	 * apply the positions from the nexus data to the host and parasite nodes.
	 */
	public int applyPositions(IPositionNode node) {
		int unpositionedNode = 0;
		for (int index = 0; index < node.getChildren().size(); index++) {
			unpositionedNode += applyPositions(node.getChildren().get(index));
		}
		Tupel<Integer, Integer> position = this.positions.get(node.getLabel());
		if (position != null) {
			node.setPosX(position.getValue1());
			node.setPosY(position.getValue2());
		} else
			unpositionedNode++;
		return unpositionedNode;
	}

	/**
	 * find the node with the given name in the host tree.
	 */
	public IPositionNode getNodeByLabel(IPositionNode node, String nodeName) {
		if (node.getLabel().equals(nodeName))
			return node;
		for (int index = 0; index < node.getChildren().size(); index++) {
			IPositionNode childNode = getNodeByLabel(
					node.getChildren().get(index), nodeName);
			if (childNode != null)
				return childNode;
		}
		return null;
	}

	/**
	 * @param label
	 * @param currentNode
	 * @return true if the label is not used by any other than the current node,
	 *         else false
	 */
	public boolean checkNodeLabel(String label, IPositionNode currentNode) {
		List<IPositionNode> nodeList = getNodeList(getHostRoot());
		nodeList.addAll(getNodeList(getParasiteRoot()));
		for (IPositionNode node : nodeList) {
			if (node.getLabel().equals(label) && !node.equals(currentNode)) {
				return false;
			}
		}
		return true;
	}

	public List<IPositionNode> getNodeList(IPositionNode node) {
		List<IPositionNode> nodeList = new ArrayList<IPositionNode>();
		nodeList.add(node);
		for (IPositionNode child : node.getChildren()) {
			nodeList.addAll(getNodeList(child));
		}
		return nodeList;
	}

	/**
	 * recalculates the number and the children count of all nodes in both trees
	 */
	public void recountAndRenumber() {
		recountAndRenumber(getHostRoot(), 0);
		recountAndRenumber(getParasiteRoot(), 0);
	}

	/**
	 * @param node
	 *            as the root node of the subtree which should be recounted and
	 *            renumbered
	 * @param number
	 *            as the starting number for this root node
	 * @return the number of children of this node
	 */
	private int recountAndRenumber(IPositionNode node, int number) {
		int nextNumber = number + 1;
		int childrenCount = 0;
		for (IPositionNode child : node.getChildren()) {
			int singleChildrenCount = recountAndRenumber(child, nextNumber);
			childrenCount += singleChildrenCount;
			nextNumber += singleChildrenCount;
		}
		node.setNumber(number);
		if (node.getParent() != null)
			node.setLevel(node.getParent().getLevel() + 1);
		else
			node.setLevel(0);
		node.setChildrenCount(childrenCount);
		return childrenCount + 1;
	}

	/**
	 * @param hostNode
	 * @param parasiteNode
	 * @return the list of association partners for the given host node in the
	 *         subtree beginning at the given parasite node
	 */
	public List<InputParasiteNode> getAssociationPartners(
			InputHostNode hostNode, InputParasiteNode parasiteNode) {
		List<InputParasiteNode> nodeList = new ArrayList<InputParasiteNode>();
		if (parasiteNode.getAssociations().contains(hostNode)) {
			nodeList.add(parasiteNode);
		}
		for (InputParasiteNode childNode : parasiteNode.getChildren()) {
			nodeList.addAll(getAssociationPartners(hostNode, childNode));
		}
		return nodeList;
	}

	/**
	 * @param hostNode
	 * @param parasiteNode
	 * @return the list of reconstructed association partners for the given host
	 *         node in the subtree beginning at the given parasite node
	 */
	public List<InputParasiteNode> getReconstructionAssociationPartners(
			InputHostNode hostNode, InputParasiteNode parasiteNode) {
		List<InputParasiteNode> nodeList = new ArrayList<InputParasiteNode>();
		if (parasiteNode.getReconstructionAssociations().contains(hostNode)) {
			nodeList.add(parasiteNode);
		}
		for (InputParasiteNode childNode : parasiteNode.getChildren()) {
			nodeList.addAll(getReconstructionAssociationPartners(hostNode,
					childNode));
		}
		return nodeList;
	}

	/**
	 * @param parasiteNode
	 */
	public void deleteNoneLeafAssociations(InputParasiteNode parasiteNode) {
		if (!parasiteNode.isLeaf()) {
			parasiteNode.getAssociations().clear();
			for (InputParasiteNode child : parasiteNode.getChildren()) {
				deleteNoneLeafAssociations(child);
			}
		}
	}

	// TODO: WARNING pruning the host tree only works correct in case of pure
	// leaf to leaf mapping. No internal node mappings are allowed.
	// Furthermore the correction of the number of sortings is only valid when
	// no complete host switches occur and if there is no delay of takeof
	// resp. landing site
	public void pruneHostTree() {

		List<InputHostNode> associatedHosts = new ArrayList<InputHostNode>();
		for (INode parasite : this.getNodeList(this.getParasiteRoot())) {
			List<InputHostNode> hosts = ((InputParasiteNode) parasite)
					.getAssociations();
			for (InputHostNode host : hosts) {
				if (!associatedHosts.contains(host)) {
					associatedHosts.add(host);
				}
			}
		}

		Map<Integer, InputHostNode> deletedHosts = new TreeMap<Integer, InputHostNode>();
		pruneHostTree(this.getHostRoot(), associatedHosts, deletedHosts);
		cleaningUpTheReconstructionInfoAfterPruningTheHostTree(deletedHosts);
		getHostRoot().renumberNodes(0);
		getHostRoot().renameNodes();

	}

	private void cleaningUpTheReconstructionInfoAfterPruningTheHostTree(
			Map<Integer, InputHostNode> deletedHosts) {
		for (InputHostNode host : deletedHosts.values()) {
			for (INode iNode : this.getNodeList(getParasiteRoot())) {
				InputParasiteNode parasite = (InputParasiteNode) iNode;
				if (parasite.getReconstructionAssociations().contains(host)) {
					parasite.getReconstructionAssociations().remove(host);
					if (!host.isLeaf()
							&& !parasite.getReconstructionAssociations()
									.contains(host.getChildren().get(0))) {
						parasite.getReconstructionAssociations().add(
								host.getChildren().get(0));
					}
				}
			}
		}

	}

	private void pruneHostTree(InputHostNode hostNode,
			List<InputHostNode> associatedHosts,
			Map<Integer, InputHostNode> deletedHosts) {
		if (!hostNode.isLeaf()) {
			List<InputHostNode> children = new ArrayList<InputHostNode>();
			for (InputHostNode child : hostNode.getChildren()) {
				children.add(child);
			}
			for (InputHostNode child : children) {
				pruneHostTree(child, associatedHosts, deletedHosts);
			}
		}
		if (hostNode.isLeaf() && !associatedHosts.contains(hostNode)) {
			InputHostNode parent = hostNode.getParent();
			if (parent != null) {
				parent.getChildren().remove(hostNode);
			}
			deletedHosts.put(hostNode.getNumber(), hostNode);
		} else if (hostNode.getDirectChildrenCount() == 1) {
			// Adjust the reconstruction events (sortings) too
			List<InputParasiteNode> reconstructionAssociatedParasites = getReconstructionAssociationPartners(
					hostNode, getParasiteRoot());
			for (InputParasiteNode parasite : reconstructionAssociatedParasites) {
				boolean foundAssociatedChild = false;
				for (InputParasiteNode associatedParasiteChild : parasite
						.getChildren()) {
					if (associatedParasiteChild.getReconstructionAssociations()
							.contains(hostNode)) {
						foundAssociatedChild = true;
						break;
					}
				}
				if (!foundAssociatedChild) {
					this.reconstructionEvents = this.reconstructionEvents
							.add(new EventType(0, -1, 0, 0));
				}
			}

			InputHostNode parent = hostNode.getParent();
			InputHostNode child = hostNode.getChildren().get(0);
			if (parent != null) {
				int index = parent.getChildren().indexOf(hostNode);
				parent.getChildren().add(index, child);
				parent.getChildren().remove(hostNode);
				child.setParent(parent);
			} else {
				this.setHostRoot(child);
				child.setParent(null);
			}
			deletedHosts.put(hostNode.getNumber(), hostNode);
		}

	}
}
