/**
 * 
 */
package com.wieseke.cptk.input.format.nexus;

import java.io.IOException;
import java.io.OutputStream;
import java.math.BigDecimal;
import java.util.ArrayList;
import java.util.List;
import java.util.Map.Entry;

import org.apache.commons.io.IOUtils;

import com.wieseke.cptk.common.api.ICophylogeny;
import com.wieseke.cptk.common.api.IHostNode;
import com.wieseke.cptk.common.api.INode;
import com.wieseke.cptk.common.api.IPositionNode;
import com.wieseke.cptk.common.api.ISerializer;
import com.wieseke.cptk.common.exceptions.SerializeException;
import com.wieseke.cptk.common.util.NullLogger;
import com.wieseke.cptk.input.constants.InputFormatNexusConstants;
import com.wieseke.cptk.input.dao.InputCophylogeny;
import com.wieseke.cptk.input.dao.InputHostNode;
import com.wieseke.cptk.input.dao.InputParasiteNode;

/**
 * @author Nicolas Wieseke
 * 
 */
public class NexusSerializer implements ISerializer {

	/**
	 * Logger
	 */
	// TODO NullLogger
	// static Logger logger = Logger.getLogger(NexusSerializer.class);
	static NullLogger logger = new NullLogger();

	public void serialize(OutputStream outputStream, ICophylogeny cophylogeny)
			throws SerializeException {
		logger.debug("---start serialize to file---");

		String nexusOutput = serialize((InputCophylogeny) cophylogeny);
		try {
			IOUtils.write(nexusOutput, outputStream);
		} catch (IOException e) {
			throw new SerializeException(
					"Error while serializing to nexus file.", e);
		}

		logger.debug("---end serialize to file---");
	}

	public String serialize(InputCophylogeny cophylogeny)
			throws SerializeException {
		logger.debug("---start serialize to string---");

		String nexusOutput = InputFormatNexusConstants.NEXUS_DIRECTIVE + "\n";
		nexusOutput += handleTaxaBlock(cophylogeny);
		nexusOutput += handleTreesBlock(cophylogeny);
		nexusOutput += handleCophylogenyBlock(cophylogeny, true);

		logger.debug("---end serialize to string---");
		return nexusOutput;
	}

	public String serialize(InputCophylogeny cophylogeny, boolean usePositions)
			throws SerializeException {
		logger.debug("---start serialize to string---");

		String nexusOutput = InputFormatNexusConstants.NEXUS_DIRECTIVE + "\n";
		nexusOutput += handleTaxaBlock(cophylogeny);
		nexusOutput += handleTreesBlock(cophylogeny);
		nexusOutput += handleCophylogenyBlock(cophylogeny, usePositions);

		logger.debug("---end serialize to string---");
		return nexusOutput;
	}

	private String handleTaxaBlock(InputCophylogeny cophylogeny) {
		String taxaBlock = InputFormatNexusConstants.BLOCK_BEGIN_COMMAND + " "
				+ InputFormatNexusConstants.TAXA_BLOCK_NAME + ";\n";
		List<String> names = new ArrayList<String>();
		names = getLeafNames(cophylogeny.getHostRoot(), names);
		names = getLeafNames(cophylogeny.getParasiteRoot(), names);
		taxaBlock += "\t"
				+ InputFormatNexusConstants.TAXA_BLOCK_DIMENSIONS_COMMAND + " "
				+ InputFormatNexusConstants.TAXA_BLOCK_DIMENSIONS_NTAX_COMMAND
				+ " = " + names.size() + ";\n";
		taxaBlock += "\t"
				+ InputFormatNexusConstants.TAXA_BLOCK_TAXLABELS_COMMAND + "\n";
		for (String name : names) {
			taxaBlock += "\t\t" + name + "\n";
		}
		taxaBlock += "\t\t;\n";
		taxaBlock += InputFormatNexusConstants.BLOCK_END_COMMAND + ";\n\n";
		return taxaBlock;
	}

	private List<String> getLeafNames(INode node, List<String> names) {
		if (node.isLeaf())
			names.add(transformLabel(node.getLabel()));
		else {
			for (INode child : node.getChildren()) {
				names = getLeafNames(child, names);
			}
		}
		return names;
	}

	private String handleTreesBlock(InputCophylogeny cophylogeny) {
		String treesBlock = InputFormatNexusConstants.BLOCK_BEGIN_COMMAND + " "
				+ InputFormatNexusConstants.TREES_BLOCK_NAME + ";\n";

		treesBlock += handleTreesBlockTranslationCommand(cophylogeny);
		treesBlock += handleTreesBlockHostTreeCommand(cophylogeny);
		treesBlock += handleTreesBlockParasiteTreeCommand(cophylogeny);

		treesBlock += InputFormatNexusConstants.BLOCK_END_COMMAND + ";\n\n";
		return treesBlock;
	}

	private String handleTreesBlockTranslationCommand(
			InputCophylogeny cophylogeny) {
		String translations = "\t"
				+ InputFormatNexusConstants.TREES_BLOCK_TRANSLATE_COMMAND
				+ "\n";
		translations += getHostNodeTranslations(cophylogeny.getHostRoot());
		translations += getParasiteNodeTranslations(cophylogeny
				.getParasiteRoot());
		if (translations.endsWith(",\n"))
			translations = translations.substring(0, translations.length() - 2)
					+ "\n";
		translations += "\t\t;\n";
		return translations;
	}

	private String getHostNodeTranslations(INode node) {
		String translations = "\t\tH" + node.getNumber() + "\t"
				+ transformLabel(node.getLabel()) + ",\n";
		for (INode child : node.getChildren()) {
			translations += getHostNodeTranslations(child);
		}
		return translations;
	}

	private String getParasiteNodeTranslations(INode node) {
		String translations = "\t\tP" + node.getNumber() + "\t"
				+ transformLabel(node.getLabel()) + ",\n";
		for (INode child : node.getChildren()) {
			translations += getParasiteNodeTranslations(child);
		}
		return translations;
	}

	private String handleTreesBlockHostTreeCommand(InputCophylogeny cophylogeny) {
		String hostTree = "\t"
				+ InputFormatNexusConstants.TREES_BLOCK_TREE_COMMAND + " "
				+ InputFormatNexusConstants.TREES_BLOCK_HOST_TREE_NAME + " = ";
		hostTree += getHostTree(cophylogeny.getHostRoot());
		hostTree += ";\n";
		return hostTree;
	}

	private String handleTreesBlockParasiteTreeCommand(
			InputCophylogeny cophylogeny) {
		String parasiteTree = "\t"
				+ InputFormatNexusConstants.TREES_BLOCK_TREE_COMMAND + " "
				+ InputFormatNexusConstants.TREES_BLOCK_PARASITE_TREE_NAME
				+ " = ";
		parasiteTree += getParasiteTree(cophylogeny.getParasiteRoot());
		parasiteTree += ";\n";
		return parasiteTree;
	}

	private String getHostTree(INode node) {
		String tree = "";
		if (node.isLeaf() || node.getChildren().size() == 0)
			tree = "H" + String.valueOf(node.getNumber());
		else {
			tree = "(";
			for (INode child : node.getChildren()) {
				tree += getHostTree(child) + ",";
			}
			tree = tree.substring(0, tree.length() - 1);
			tree += ")H" + node.getNumber();
		}
		return tree;
	}

	private String getParasiteTree(INode node) {
		String tree = "";
		if (node.isLeaf() || node.getChildren().size() == 0)
			tree = "P" + String.valueOf(node.getNumber());
		else {
			tree = "(";
			for (INode child : node.getChildren()) {
				tree += getParasiteTree(child) + ",";
			}
			tree = tree.substring(0, tree.length() - 1);
			tree += ")P" + node.getNumber();
		}
		return tree;
	}

	private String handleCophylogenyBlock(InputCophylogeny cophylogeny,
			boolean usePositions) {
		String cophylogenyBlock = InputFormatNexusConstants.BLOCK_BEGIN_COMMAND
				+ " " + InputFormatNexusConstants.COPHYLOGENY_BLOCK_NAME
				+ ";\n";

		cophylogenyBlock += handleCophylogenyBlockRanksCommand(cophylogeny);
		cophylogenyBlock += handleCophylogenyBlockPhiCommand(cophylogeny);
		cophylogenyBlock += handleCophylogenyBlockReconstructionEventsCommand(cophylogeny);
		cophylogenyBlock += handleCophylogenyBlockReconstructionCommand(cophylogeny);
		if (usePositions)
			cophylogenyBlock += handleCophylogenyBlockPositionsCommand(cophylogeny);
		cophylogenyBlock += handleCophylogenyBlockCostsCommand(cophylogeny);
		cophylogenyBlock += handleCophylogenyBlockOptionsCommand(cophylogeny);

		cophylogenyBlock += InputFormatNexusConstants.BLOCK_END_COMMAND
				+ ";\n\n";
		return cophylogenyBlock;
	}

	private String handleCophylogenyBlockRanksCommand(
			InputCophylogeny cophylogeny) {
		String ranks = "[RANKS represents the ranks of the nodes in the tree]\n[Syntax is: nodename timezone_from timezone_to]\n";
		ranks += "\t"
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_RANKS_COMMAND
				+ "\n";
		ranks += getHostNodeRanks((InputHostNode) cophylogeny.getHostRoot());
		ranks += getParasiteNodeRanks((InputParasiteNode) cophylogeny
				.getParasiteRoot());
		if (ranks.endsWith(",\n"))
			ranks = ranks.substring(0, ranks.length() - 2) + "\n";
		ranks += "\t\t;\n";
		return ranks;
	}

	private String getHostNodeRanks(InputHostNode node) {

		String ranks = "\t\t" + transformLabel(node.getLabel()) + "\t"
				+ node.getRank() + "\t" + node.getRank() + ",\n";
		for (INode child : node.getChildren()) {
			ranks += getHostNodeRanks((InputHostNode) child);
		}
		return ranks;
	}

	private String getParasiteNodeRanks(InputParasiteNode node) {

		String ranks = "\t\t" + transformLabel(node.getLabel()) + "\t"
				+ node.getRank().getRankFrom() + "\t"
				+ node.getRank().getRankTo() + ",\n";
		for (INode child : node.getChildren()) {
			ranks += getParasiteNodeRanks((InputParasiteNode) child);
		}
		return ranks;
	}

	private String handleCophylogenyBlockPhiCommand(InputCophylogeny cophylogeny) {
		String phi = "[PHI represents the associations from the parasite leaf nodes to the host leaf nodes]\n[Syntax is: parasite_leaf_name host_leaf_name]\n";
		phi += "\t" + InputFormatNexusConstants.COPHYLOGENY_BLOCK_PHI_COMMAND
				+ "\n";

		phi += getAssociations((InputParasiteNode) cophylogeny
				.getParasiteRoot());
		if (phi.endsWith(",\n"))
			phi = phi.substring(0, phi.length() - 2) + "\n";
		phi += "\t\t;\n";
		return phi;
	}

	private String getAssociations(InputParasiteNode node) {
		String associations = "";
		for (IHostNode hostNode : node.getAssociations()) {
			associations += "\t\t" + transformLabel(node.getLabel()) + "\t"
					+ transformLabel(hostNode.getLabel()) + ",\n";
		}
		for (INode child : node.getChildren()) {
			associations += getAssociations((InputParasiteNode) child);
		}
		return associations;
	}

	private String handleCophylogenyBlockReconstructionEventsCommand(
			InputCophylogeny cophylogeny) {
		String events = "[RECONSTRUCTIONEVENTS represents the events which occurred in the reconstruction. It includes 'COSPECIATION', 'DUPLICATION', 'EXTINCTION', 'SORTING' and 'HOSTSWITCH']\n[Syntax is: operation cost]\n";
		events += "\t"
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_RECONSTRUCTIONEVENTS_COMMAND
				+ "\n";
		events += getEvents(cophylogeny);
		if (events.endsWith(",\n"))
			events = events.substring(0, events.length() - 2) + "\n";
		events += "\t\t;\n";
		return events;
	}

	private String getEvents(InputCophylogeny cophylogeny) {
		String events = "";
		for (Entry<String, Integer> event : cophylogeny
				.getReconstructionEvents().getEvents().entrySet()) {
			events += "\t\t" + event.getKey() + "\t" + event.getValue() + ",\n";
		}
		return events;
	}

	private String handleCophylogenyBlockReconstructionCommand(
			InputCophylogeny cophylogeny) {
		String reconstruction = "[RECONSTRUCTION represents the associations from the parasite nodes to the host nodes as they occured in the reconstruction/simulation]\n[Syntax is: parasite_leaf_name host_leaf_name]\n";
		reconstruction += "\t"
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_RECONSTRUCTION_COMMAND
				+ "\n";

		reconstruction += getReconstructionAssociations((InputParasiteNode) cophylogeny
				.getParasiteRoot());
		if (reconstruction.endsWith(",\n"))
			reconstruction = reconstruction.substring(0,
					reconstruction.length() - 2)
					+ "\n";
		reconstruction += "\t\t;\n";
		return reconstruction;
	}

	private String getReconstructionAssociations(InputParasiteNode node) {
		String associations = "";
		for (IHostNode hostNode : node.getReconstructionAssociations()) {
			associations += "\t\t" + transformLabel(node.getLabel()) + "\t"
					+ transformLabel(hostNode.getLabel()) + ",\n";
		}
		for (INode child : node.getChildren()) {
			associations += getReconstructionAssociations((InputParasiteNode) child);
		}
		return associations;
	}

	private String handleCophylogenyBlockPositionsCommand(
			InputCophylogeny cophylogeny) {
		String positions = "[POSITIONS represents the x and y position of the node]\n[Syntax is: nodename x_position y_position]\n";
		positions += "\t"
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_POSITIONS_COMMAND
				+ "\n";
		positions += getNodePositions(cophylogeny.getHostRoot());
		positions += getNodePositions(cophylogeny.getParasiteRoot());
		if (positions.endsWith(",\n"))
			positions = positions.substring(0, positions.length() - 2) + "\n";
		positions += "\t\t;\n";
		return positions;
	}

	private String getNodePositions(IPositionNode node) {
		String positions = "\t\t" + transformLabel(node.getLabel()) + "\t"
				+ node.getPosX() + "\t" + node.getPosY() + ",\n";
		for (IPositionNode child : node.getChildren()) {
			positions += getNodePositions(child);
		}
		return positions;
	}

	private String handleCophylogenyBlockCostsCommand(
			InputCophylogeny cophylogeny) {
		String costs = "[COSTS represents the cost table for the operations 'COSPECIATION', 'DUPLICATION', 'EXTINCTION', 'SORTING' and 'HOSTSWITCH']\n[Syntax is: operation cost]\n";
		costs += "\t"
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_COSTS_COMMAND
				+ "\n";
		costs += getCosts(cophylogeny);
		if (costs.endsWith(",\n"))
			costs = costs.substring(0, costs.length() - 2) + "\n";
		costs += "\t\t;\n";
		return costs;
	}

	private String getCosts(InputCophylogeny cophylogeny) {
		String costs = "";
		for (Entry<String, BigDecimal> cost : cophylogeny.getCosts().entrySet()) {
			costs += "\t\t" + cost.getKey() + "\t" + cost.getValue() + ",\n";
		}
		return costs;
	}

	private String handleCophylogenyBlockOptionsCommand(
			InputCophylogeny cophylogeny) {
		String options = "[OPTIONS represents the options for calculating the reconstruction]\n[Syntax is: option value]\n";
		options += "\t"
				+ InputFormatNexusConstants.COPHYLOGENY_BLOCK_OPTIONS_COMMAND
				+ "\n";
		options += getOptions(cophylogeny);
		if (options.endsWith(",\n"))
			options = options.substring(0, options.length() - 2) + "\n";
		options += "\t\t;\n";
		return options;
	}

	private String getOptions(InputCophylogeny cophylogeny) {
		String options = "";
		for (Entry<String, String> option : cophylogeny.getOptions().entrySet()) {
			options += "\t\t" + option.getKey() + "\t" + option.getValue()
					+ ",\n";
		}
		return options;
	}

	private String transformLabel(String label) {
		if (label.contains(" ") || label.contains("'") || label.contains("\t")) {
			label = "'" + label.replace("'", "''") + "'";
		}
		return label;
	}
}
