/**
 * 
 */
package com.wieseke.cptk.main;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.OutputStreamWriter;
import java.util.Random;

import org.apache.commons.io.FileUtils;

import com.wieseke.cptk.common.exceptions.SerializeException;
import com.wieseke.cptk.common.log.CophylogenyLogger;
import com.wieseke.cptk.input.dao.InputCophylogeny;
import com.wieseke.cptk.input.format.nexus.NexusDeserializer;
import com.wieseke.cptk.input.format.nexus.NexusSerializer;
import com.wieseke.cptk.reconstruction.constants.StatisticConstants;
import com.wieseke.cptk.reconstruction.dao.ReconstructionCophylogeny;
import com.wieseke.cptk.statistics.TreeGenerator;

/**
 * @author Nicolas Wieseke
 * 
 */
public class Statistics {

	/**
	 * @param args
	 */
	public static void main(String[] args) {
		if (args.length == 0 || args[0].equalsIgnoreCase("-h")
				|| args[0].equalsIgnoreCase("-help")) {
			printHelp();
			return;
		}
		try {
			String filename = args[0];
			int tree = 3;
			String model = StatisticConstants.MODEL_AGE;
			int cycles = 100;
			boolean original = false;
			for (int i = 1; i < args.length; i++) {
				if (args[i].startsWith("-r") || args[i].startsWith("-R")) {
					tree = Integer.parseInt(args[i].substring(2));
				}
				if (args[i].startsWith("-m") || args[i].startsWith("-M")) {
					int arg = Integer.parseInt(args[i].substring(2));
					switch (arg) {
					case 0:
						model = StatisticConstants.MODEL_MARKOV;
						break;
					case 1:
						model = StatisticConstants.MODEL_ALDOUS;
						break;
					case 2:
						model = StatisticConstants.MODEL_PDA;
						break;
					case 3:
						model = StatisticConstants.MODEL_AGE;
						break;
					default:
						model = StatisticConstants.MODEL_AGE;
						break;
					}
				}
				if (args[i].startsWith("-c") || args[i].startsWith("-C")) {
					cycles = Integer.parseInt(args[i].substring(2));
				}
				if (args[i].startsWith("-o") || args[i].startsWith("-O")) {
					original = true;
				}
			}

			NexusDeserializer nd = new NexusDeserializer();
			InputCophylogeny cophylogeny = nd.deserialize(new FileInputStream(
					filename), filename);
			ReconstructionCophylogeny reconstructionCophylogeny = new ReconstructionCophylogeny(
					cophylogeny, cophylogeny.getCosts(),
					cophylogeny.getOptions());
			if (original) {
				OutputStreamWriter osw = new OutputStreamWriter(
						FileUtils.openOutputStream(new File(filename + ".log")));
				osw.write("Counter\tQuality\tTotalCost\tTotalEvents\tCoCost\tCoEvents\tCoProb\tCoFreq\tSoCost\tSoEvents\tSoProb\tSoFreq\tDuCost\tDuEvents\tDuProb\tDuFreq\tHsCost\tHsEvents\tHsProb\tHsFreq\tRootMapping\n");
				reconstructionCophylogeny.compute(osw);
				osw.flush();
				osw.close();
			}
			OutputStreamWriter osw = new OutputStreamWriter(
					FileUtils.openOutputStream(new File(filename + ".rnd.log")));
			osw.write("RandomCycle\tCounter\tQuality\tTotalCost\tTotalEvents\tCoCost\tCoEvents\tCoProb\tCoFreq\tSoCost\tSoEvents\tSoProb\tSoFreq\tDuCost\tDuEvents\tDuProb\tDuFreq\tHsCost\tHsEvents\tHsProb\tHsFreq\tRootMapping\n");
			for (int i = 0; i < cycles; i++) {
				InputCophylogeny cp = new TreeGenerator(new Random())
						.getRandomInputCophylogeny(filename + i, cophylogeny,
								model, tree);
				try {
					OutputStreamWriter ow = new OutputStreamWriter(
							FileUtils.openOutputStream(new File(filename + i
									+ ".nex")));
					NexusSerializer ns = new NexusSerializer();
					ow.write(ns.serialize(cp, false));
					ow.flush();
					ow.close();
				} catch (SerializeException e) {
					CophylogenyLogger.logError(e);
				} catch (IOException e) {
					CophylogenyLogger.logError(e);
				}
				reconstructionCophylogeny = new ReconstructionCophylogeny(cp,
						cophylogeny.getCosts(), cophylogeny.getOptions());
				osw.write(i + "\t");
				reconstructionCophylogeny.compute(osw);
				osw.flush();
			}
			osw.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	/**
	 * prints the help information
	 * */
	static public void printHelp() {
		System.out
				.println("\nSyntax:\n"
						+ "java -cp .:./* com.wieseke.cptk.main.Reconstruction [filename.nex]"
						+ "\nor\n"
						+ "java -cp .:./* com.wieseke.cptk.main.Statistics [filename.nex] [options]"
						+ "\n\nOptions:\n"
						+ "-rX     \tSpecify the randomisation method.\n"
						+ "        \tX=0: compute random host and parasite tree and use original leaf associations\n"
						+ "        \tX=1: compute random host tree and use original leaf associations\n"
						+ "        \tX=2: compute random parasite tree and use original leaf associations\n"
						+ "        \tX=3: use original tree and compute random leaf associations [default]\n"
						+ "-mX     \tSpecify the tree model.\n"
						+ "        \tX=0: Markov model (beta = 0)\n"
						+ "        \tX=1: Aldous model (beta = -1)\n"
						+ "        \tX=2: PDA model (beta = -1.5)\n"
						+ "        \tX=3: Age model [default]\n"
						+ "-cX     \tSpecify the number of random cycles.\n"
						+ "        \tX: a numeric value specifying the number of random cycles [default=100]\n"
						+ "-o      \tCompute the reconstruction for the original data as well.\n"
						+ "\n\nFor the java -cp parameter use ';' as classpath separanor on Windows machines (not ':').");
	}
}
