package com.wieseke.cptk.output.editor;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.jface.action.ActionContributionItem;
import org.eclipse.jface.action.IMenuManager;
import org.eclipse.jface.action.IToolBarManager;
import org.eclipse.jface.action.MenuManager;
import org.eclipse.jface.action.Separator;
import org.eclipse.ui.IActionBars;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.IWorkbenchActionConstants;
import org.eclipse.ui.IWorkbenchPage;
import org.eclipse.ui.actions.ActionFactory;
import org.eclipse.ui.part.EditorActionBarContributor;

import com.wieseke.cptk.common.action.CommonAction;
import com.wieseke.cptk.common.action.CophylogenyViewerActionFactory;
import com.wieseke.cptk.common.action.PrintAction;
import com.wieseke.cptk.output.action.OutputAction;
import com.wieseke.cptk.output.action.OutputCophylogenyViewerActionFactory;
import com.wieseke.cptk.output.action.SelectEditModeSelectionAction;

/**
 * Manages the installation/deinstallation of global actions for multi-page
 * editors. Responsible for the redirection of global actions to the active
 * editor. Multi-page contributor replaces the contributors for the individual
 * editors in the multi-page editor.
 */
public class VisualOutputEditorActionBarContributor extends
		EditorActionBarContributor {

	private List<CommonAction> exportActions = new ArrayList<CommonAction>();
	private List<CommonAction> zoomActions = new ArrayList<CommonAction>();
	SelectEditModeSelectionAction selectEditModeSelectionAction;
	private List<OutputAction> selectEditModeActions = new ArrayList<OutputAction>();
	private List<CommonAction> nodeActions = new ArrayList<CommonAction>();
	private List<CommonAction> treeActions = new ArrayList<CommonAction>();
	private List<OutputAction> reconstructionActions = new ArrayList<OutputAction>();
	private PrintAction printAction = new PrintAction("Print...");

	/**
	 * Creates a multi-page contributor.
	 */
	public VisualOutputEditorActionBarContributor() {
		super();
		createActions();
	}

	
	public void init(IActionBars bars, IWorkbenchPage page) {
		super.init(bars, page);
		initActions();
	}

	private void createActions() {
		CophylogenyViewerActionFactory exportActionsToCreate[] = new CophylogenyViewerActionFactory[] {
				CophylogenyViewerActionFactory.PRINT,
				CophylogenyViewerActionFactory.EXPORT_SVG };
		CophylogenyViewerActionFactory zoomActionsToCreate[] = new CophylogenyViewerActionFactory[] {
				CophylogenyViewerActionFactory.ZOOM_RESET,
				CophylogenyViewerActionFactory.ZOOM_IN,
				CophylogenyViewerActionFactory.ZOOM_OUT };
		OutputCophylogenyViewerActionFactory selectEditModeActionsToCreate[] = new OutputCophylogenyViewerActionFactory[] {
				OutputCophylogenyViewerActionFactory.SELECT_FOLD_NODE,
				OutputCophylogenyViewerActionFactory.SELECT_CHANGE_LABEL };
		CophylogenyViewerActionFactory nodeActionsToCreate[] = new CophylogenyViewerActionFactory[] {
				// CophylogenyViewerActionFactory.ADJUST_CHILD_NODES,
				CophylogenyViewerActionFactory.SHOW_HOST_LABELS,
				CophylogenyViewerActionFactory.SHOW_PARASITE_LABELS,
				CophylogenyViewerActionFactory.SHOW_RANK_LABELS
		// ,CophylogenyViewerActionFactory.RECALCULATE_NODE_POSITIONS
		};
		CophylogenyViewerActionFactory treeActionsToCreate[] = new CophylogenyViewerActionFactory[] {
				CophylogenyViewerActionFactory.FLIP_HORIZONTAL,
				CophylogenyViewerActionFactory.FLIP_VERTICAL,
				CophylogenyViewerActionFactory.ROTATE_LEFT,
				CophylogenyViewerActionFactory.ROTATE_RIGHT };
		OutputCophylogenyViewerActionFactory reconstructionActionsToCreate[] = new OutputCophylogenyViewerActionFactory[] {
				OutputCophylogenyViewerActionFactory.RECONSTRUCTION_PREVIOUS,
				OutputCophylogenyViewerActionFactory.RECONSTRUCTION_NEXT,
				OutputCophylogenyViewerActionFactory.RECONSTRUCTION_CHOOSE };

		for (CophylogenyViewerActionFactory factory : exportActionsToCreate) {
			CommonAction action = factory.create();
			this.exportActions.add(action);
		}
		for (CophylogenyViewerActionFactory factory : zoomActionsToCreate) {
			CommonAction action = factory.create();
			this.zoomActions.add(action);
		}
		for (OutputCophylogenyViewerActionFactory factory : selectEditModeActionsToCreate) {
			OutputAction action = factory.create();
			this.selectEditModeActions.add(action);
		}
		for (CophylogenyViewerActionFactory factory : nodeActionsToCreate) {
			CommonAction action = factory.create();
			this.nodeActions.add(action);
		}
		for (CophylogenyViewerActionFactory factory : treeActionsToCreate) {
			CommonAction action = factory.create();
			this.treeActions.add(action);
		}
		for (OutputCophylogenyViewerActionFactory factory : reconstructionActionsToCreate) {
			OutputAction action = factory.create();
			this.reconstructionActions.add(action);
		}
	}

	private void initActions() {
		for (CommonAction action : this.exportActions) {
			getPage().addSelectionListener(action);
		}
		for (CommonAction action : this.zoomActions) {
			getPage().addSelectionListener(action);
		}
		for (OutputAction action : this.selectEditModeActions) {
			getPage().addSelectionListener(action);
		}
		for (CommonAction action : this.nodeActions) {
			getPage().addSelectionListener(action);
		}
		for (CommonAction action : this.treeActions) {
			getPage().addSelectionListener(action);
		}
		for (OutputAction action : this.reconstructionActions) {
			getPage().addSelectionListener(action);
		}
		getPage().addSelectionListener(this.selectEditModeSelectionAction);
		getPage().addSelectionListener(this.printAction);
	}

	
	public void setActiveEditor(IEditorPart targetEditor) {
		super.setActiveEditor(targetEditor);
		IActionBars actionBars = getActionBars();
		if (actionBars != null) {
			IEditorPart part = getPage().getActiveEditor();
			if (!(part instanceof VisualOutputEditor)) {
				disableOutputActions();
			} else {
				enableOutputActions();
				actionBars.setGlobalActionHandler(ActionFactory.PRINT.getId(),
						printAction);
			}
			actionBars.updateActionBars();
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.ui.part.EditorActionBarContributor#contributeToMenu(org.eclipse
	 * .jface.action.IMenuManager)
	 */
	
	public void contributeToMenu(IMenuManager manager) {
		IMenuManager menu = new MenuManager("&Cophylogeny");
		manager.prependToGroup(IWorkbenchActionConstants.MB_ADDITIONS, menu);
		for (CommonAction action : this.exportActions) {
			menu.add(action);
		}
		menu.add(new Separator());
		for (CommonAction action : this.zoomActions) {
			menu.add(action);
		}
		menu.add(new Separator());
		IMenuManager selectEditModeActionMenu = new MenuManager(
				"Select &edit mode");
		for (OutputAction action : this.selectEditModeActions) {
			selectEditModeActionMenu.add(action);
		}
		menu.add(selectEditModeActionMenu);
		menu.add(new Separator());
		for (CommonAction action : this.nodeActions) {
			menu.add(action);
		}
		menu.add(new Separator());
		for (CommonAction action : this.treeActions) {
			menu.add(action);
		}
		menu.add(new Separator());
		for (OutputAction action : this.reconstructionActions) {
			menu.add(action);
		}
		menu.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see
	 * org.eclipse.ui.part.EditorActionBarContributor#contributeToToolBar(org
	 * .eclipse.jface.action.IToolBarManager)
	 */
	
	public void contributeToToolBar(IToolBarManager manager) {
		manager.add(new Separator());
		for (CommonAction action : this.exportActions) {
			manager.add(action);
		}
		manager.add(new Separator());
		for (CommonAction action : this.zoomActions) {
			manager.add(action);
		}
		manager.add(new Separator());
		selectEditModeSelectionAction = new SelectEditModeSelectionAction(
				"Select edit mode");
		for (OutputAction action : this.selectEditModeActions) {
			selectEditModeSelectionAction.addActionToMenu(action);
		}
		ActionContributionItem selectEditModeSelectionItem = new ActionContributionItem(
				selectEditModeSelectionAction);
		selectEditModeSelectionItem
				.setMode(ActionContributionItem.MODE_FORCE_TEXT);
		manager.add(selectEditModeSelectionItem);
		manager.add(new Separator());
		for (CommonAction action : this.nodeActions) {
			manager.add(action);
		}
		manager.add(new Separator());
		for (CommonAction action : this.treeActions) {
			manager.add(action);
		}
		manager.add(new Separator());
		for (OutputAction action : this.reconstructionActions) {
			manager.add(action);
		}
		manager.add(new Separator(IWorkbenchActionConstants.MB_ADDITIONS));

	}

	private void enableOutputActions() {
		for (CommonAction action : this.exportActions) {
			action.setEnabled(true);
		}
		for (CommonAction action : this.zoomActions) {
			action.setEnabled(true);
		}
		for (OutputAction action : this.selectEditModeActions) {
			action.setEnabled(true);
		}
		for (CommonAction action : this.nodeActions) {
			action.setEnabled(true);
		}
		for (CommonAction action : this.treeActions) {
			action.setEnabled(true);
		}
		for (OutputAction action : this.reconstructionActions) {
			action.setEnabled(true);
		}
		selectEditModeSelectionAction.setEnabled(true);
		printAction.setEnabled(true);
	}

	private void disableOutputActions() {
		for (CommonAction action : this.exportActions) {
			action.setEnabled(false);
		}
		for (CommonAction action : this.zoomActions) {
			action.setEnabled(false);
		}
		for (OutputAction action : this.selectEditModeActions) {
			action.setEnabled(false);
		}
		for (CommonAction action : this.nodeActions) {
			action.setEnabled(false);
		}
		for (CommonAction action : this.treeActions) {
			action.setEnabled(false);
		}
		for (OutputAction action : this.reconstructionActions) {
			action.setEnabled(false);
		}
		selectEditModeSelectionAction.setEnabled(false);
		printAction.setEnabled(false);
	}

}
